package net.knarcraft.bookswithoutborders;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.bukkit.ChatColor;
import org.bukkit.Material;
import org.bukkit.Tag;
import org.bukkit.block.Sign;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.Action;
import org.bukkit.event.block.SignChangeEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.bukkit.event.player.PlayerItemHeldEvent;
import org.bukkit.event.player.PlayerJoinEvent;
import org.bukkit.inventory.EquipmentSlot;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.inventory.meta.BookMeta;
import org.bukkit.inventory.meta.ItemMeta;

/**
 * A listener for relevant events
 */
public class BooksWithoutBordersListener implements Listener {

    private final String slash = BooksWithoutBorders.SLASH;
    private final String bookFolder = BooksWithoutBorders.bwb.getDataFolder().getAbsolutePath() + slash + "Books" + slash;

    /**
     * Updates old books to a newer format
     *
     * @param player  <p>The player holding the book</p>
     * @param oldBook <p>Metadata about the held book</p>
     * @return <p>An updated book</p>
     */
    public ItemStack updateBook(Player player, BookMeta oldBook) {
        //handles hacked title-less books
        if (oldBook.getTitle() == null || oldBook.getTitle().length() < 3) {
            return null;
        }

        if (oldBook.getTitle().substring(oldBook.getTitle().length() - 3).equalsIgnoreCase("[U]")) {
            String fileName;

            if (oldBook.getAuthor() != null && oldBook.getAuthor().equalsIgnoreCase("unknown")) {
                //Unknown author is ignored
                fileName = oldBook.getTitle();
            } else {
                fileName = oldBook.getTitle() + BooksWithoutBorders.titleAuthorSeparator + oldBook.getAuthor();
            }

            String cleanPlayerName = BooksWithoutBorders.bwb.cleanString(player.getName());

            String[] possiblePaths = new String[]{
                    bookFolder + fileName + ".yml",
                    bookFolder + fileName + ".txt",
                    bookFolder + cleanPlayerName + slash + fileName + ".yml",
                    bookFolder + cleanPlayerName + slash + fileName + ".txt"
            };

            for (String path : possiblePaths) {
                File file = new File(path);
                if (file.isFile()) {
                    return BooksWithoutBorders.bwb.loadBook(player, fileName, "true", "player");
                }
            }
            return null;
        }

        return null;
    }

    @EventHandler
    public void onHold(PlayerItemHeldEvent e) {
        Player player = e.getPlayer();
        int selectedSlot = e.getNewSlot();
        PlayerInventory playerInventory = player.getInventory();
        ItemStack selectedItem = playerInventory.getItem(selectedSlot);

        //Ignore irrelevant items
        if (selectedItem == null || selectedItem.getType() != Material.WRITTEN_BOOK) {
            return;
        }

        ItemMeta itemMetadata = selectedItem.getItemMeta();
        if (itemMetadata == null) {
            return;
        }

        //Update the book the user is viewing
        updateBookInHand(player, itemMetadata, true);
    }

    @EventHandler
    public void onPlayerJoin(PlayerJoinEvent e) {
        //Handle new players
        if (!BooksWithoutBorders.bwb.hasPlayedBefore(e.getPlayer().getName())) {
            Player player = e.getPlayer();
            boolean sendMessage = true;

            //Gives new players necessary books
            for (String bookName : BooksWithoutBorders.firstBooks) {
                sendMessage = giveBookToNewPlayer(bookName, player, sendMessage);
            }
        }

        //Updates any books in either hand
        ItemStack mainHandItem = e.getPlayer().getInventory().getItemInMainHand();
        ItemStack offHandItem = e.getPlayer().getInventory().getItemInOffHand();
        if (mainHandItem.getType() == Material.WRITTEN_BOOK) {
            ItemMeta itemMetadata = mainHandItem.getItemMeta();
            updateBookInHand(e.getPlayer(), itemMetadata, true);
        }
        if (offHandItem.getType() == Material.WRITTEN_BOOK) {
            ItemMeta itemMetadata = offHandItem.getItemMeta();
            updateBookInHand(e.getPlayer(), itemMetadata, false);
        }
    }

    /**
     * Gives a book to a player joining for the first time
     * @param bookName <p>The name of the book to give</p>
     * @param player <p>The player which just joined</p>
     * @param sendMessage <p>Whether to send a message to the joining player</p>
     * @return <p>True if a message has yet to be sent</p>
     */
    private boolean giveBookToNewPlayer(String bookName, Player player, boolean sendMessage) {

        if (!bookName.equalsIgnoreCase(" ")) {
            //handles loadlist numbers
            for (int x = 0; x < bookName.length(); x++) {
                if (!Character.isDigit(bookName.charAt(x))) {
                    break;
                }
                if (x == bookName.length() - 1) {
                    BooksWithoutBorders.loadList.put(player.getName(), BooksWithoutBorders.bwb.listFiles(player, true, 0, true));
                }
            }


            ItemStack newBook = BooksWithoutBorders.bwb.loadBook(player, bookName, "true", "public");
            if (newBook != null) {
                player.getInventory().addItem(newBook);
            }

            if (!BooksWithoutBorders.fBMessage.equalsIgnoreCase(" ") && newBook != null && sendMessage) {
                sendMessage = false;
                final Player fp = player;
                BooksWithoutBorders.bwb.getServer().getScheduler().scheduleSyncDelayedTask(BooksWithoutBorders.bwb,
                        () -> fp.sendMessage(BooksWithoutBorders.fBMessage), 40L);
            }
        }
        return sendMessage;
    }

    /**
     * Updates a book in one of the player's hands
     *
     * @param player       <p>The player to update</p>
     * @param itemMetadata <p>Information about the held book</p>
     * @param mainHand     <p>Whether to update the book in the player's main hand</p>
     */
    private void updateBookInHand(Player player, ItemMeta itemMetadata, boolean mainHand) {
        PlayerInventory playerInventory = player.getInventory();
        ItemStack updatedBook = updateBook(player, (BookMeta) itemMetadata);
        if (updatedBook != null) {
            if (mainHand) {
                playerInventory.setItemInMainHand(updatedBook);
            } else {
                playerInventory.setItemInOffHand(updatedBook);
            }
        }
    }

    @EventHandler
    public void onSignChange(SignChangeEvent e) {
        if (e.getLine(0).equalsIgnoreCase("[BwB]") && e.getPlayer().hasPermission("bookswithoutborders.signs")) {
            e.setLine(0, ChatColor.DARK_GREEN + "[BwB]");

            if ((e.getLine(1).equalsIgnoreCase("[Encrypt]") || e.getLine(1).equalsIgnoreCase("[Decrypt]") || e.getLine(1).equalsIgnoreCase("[Give]")) && e.getLine(2).replace(" ", "").length() > 0) {
                e.setLine(1, ChatColor.DARK_BLUE + e.getLine(1));
                if (e.getLine(1).equalsIgnoreCase(ChatColor.DARK_BLUE + "[Encrypt]") || e.getLine(1).equalsIgnoreCase(ChatColor.DARK_BLUE + "[Decrypt]")) {
                    e.setLine(2, ChatColor.MAGIC + e.getLine(2));
                    e.setLine(3, ChatColor.DARK_BLUE + e.getLine(3));
                }

                if (e.getLine(1).equalsIgnoreCase(ChatColor.DARK_BLUE + "[Give]")) {
                    if (e.getLine(2).length() > 13 || e.getLine(3).length() > 13) {
                        e.getPlayer().sendMessage(ChatColor.RED + "[Give] signs' 3rd and 4th lines must be 13 characters or less!");
                        e.setLine(2, ChatColor.DARK_RED + e.getLine(2));
                        e.setLine(3, ChatColor.DARK_RED + e.getLine(3));
                        return;
                    }

                    //Tests if a full file name has been supplied
                    if ((new File(BooksWithoutBorders.bwb.getDataFolder().getAbsolutePath() + BooksWithoutBorders.SLASH + "Books" + BooksWithoutBorders.SLASH + e.getLine(2) + e.getLine(3)).isFile()) ||
                            (new File(BooksWithoutBorders.bwb.getDataFolder().getAbsolutePath() + BooksWithoutBorders.SLASH + "Books" + BooksWithoutBorders.SLASH + e.getLine(2) + e.getLine(3) + ".txt").isFile()) ||
                            (new File(BooksWithoutBorders.bwb.getDataFolder().getAbsolutePath() + BooksWithoutBorders.SLASH + "Books" + BooksWithoutBorders.SLASH + e.getLine(2) + e.getLine(3) + ".yml").isFile())) {
                        e.setLine(2, ChatColor.DARK_GREEN + e.getLine(2));
                        e.setLine(3, ChatColor.DARK_GREEN + e.getLine(3));
                    }

                    //Tests if a load list number has been supplied
                    else {
                        File dirFile = new File(BooksWithoutBorders.bwb.getDataFolder().getAbsolutePath() + BooksWithoutBorders.SLASH + "Books");
                        for (int x = 0; x < e.getLine(2).length(); x++) {
                            if (!Character.isDigit(e.getLine(2).charAt(x))) {
                                e.setLine(2, ChatColor.DARK_RED + e.getLine(2));
                                e.setLine(3, ChatColor.DARK_RED + e.getLine(3));
                                break;
                            }
                            if (x == e.getLine(2).length() - 1) {
                                List<String> fList = new ArrayList<>();
                                for (int y = 0; y < dirFile.list().length; y++) {
                                    if (dirFile.listFiles()[y].isFile()) {
                                        fList.add(dirFile.listFiles()[y].getName());
                                    }
                                }

                                try {
                                    if (Integer.parseInt(e.getLine(2)) >= 0 && Integer.parseInt(e.getLine(2)) <= fList.size()) {
                                        BooksWithoutBorders.loadList.put(e.getPlayer().getName(), BooksWithoutBorders.bwb.listFiles(e.getPlayer(), true, 0, true));
                                        e.setLine(2, ChatColor.DARK_GREEN + e.getLine(2));
                                        e.setLine(3, ChatColor.DARK_GREEN + e.getLine(3));
                                    } else {
                                        e.setLine(2, ChatColor.DARK_RED + e.getLine(2));
                                        e.setLine(3, ChatColor.DARK_RED + e.getLine(3));
                                    }
                                } catch (NumberFormatException ignored) {
                                }
                            }
                        }
                    }
                }
            } else {
                e.setLine(1, ChatColor.DARK_RED + e.getLine(1));
            }
        }
    }

    @EventHandler
    public void onClick(PlayerInteractEvent e) {
        if (e.getClickedBlock() == null) {
            return;
        }

        Material clickedBlockType = e.getClickedBlock().getType();
        Player player = e.getPlayer();
        PlayerInventory playerInventory = player.getInventory();
        EquipmentSlot hand = e.getHand();
        if (hand == null) {
            return;
        }
        ItemStack heldItem = playerInventory.getItem(hand);
        Material heldItemType = heldItem.getType();

        if (e.getAction() == Action.RIGHT_CLICK_BLOCK && (Tag.SIGNS.isTagged(clickedBlockType) || Tag.WALL_SIGNS.isTagged(clickedBlockType))) {
            //The player right-clicked a sign
            Sign sign = (Sign) e.getClickedBlock().getState();
            if (signLineEquals(sign, 0, "[BooksWithoutBorders]", ChatColor.DARK_GREEN)) {

                if (signLineEquals(sign, 1, "[Encrypt]", ChatColor.DARK_BLUE)) {
                    encryptHeldBookUsingSign(sign, heldItemType, player, hand);
                } else if (signLineEquals(sign, 1, "[Decrypt]", ChatColor.DARK_BLUE)) {
                    decryptHeldBookUsingSign(sign, heldItemType, player, hand);
                } else if (signLineEquals(sign, 1, "[Give]", ChatColor.DARK_BLUE) &&
                        getSignLineColor(sign, 2) == ChatColor.DARK_GREEN) {
                    giveBook(sign, player);
                }
            }
        } else if (heldItemType == Material.WRITTEN_BOOK && (e.getAction() == Action.LEFT_CLICK_AIR
                || e.getAction() == Action.LEFT_CLICK_BLOCK)) {
            BookMeta oldBook = (BookMeta) heldItem.getItemMeta();
            if (oldBook == null) {
                return;
            }

            decryptBook(oldBook, player, heldItem, hand);
        }
    }

    /**
     * Decrypts a book decryptable for a given group
     *
     * @param oldBook  <p>Metadata for the encrypted book</p>
     * @param player   <p>The player decrypting the book</p>
     * @param heldItem <p>The type of the held book</p>
     * @param hand     <p>The hand the player is using to hold the book</p>
     */
    private void decryptBook(BookMeta oldBook, Player player, ItemStack heldItem, EquipmentSlot hand) {
        ItemStack newBook;

        //Check if the book is encrypted by Books Without Borders
        if (!oldBook.hasLore() || oldBook.getLore() == null || !oldBook.getLore().get(0).contains(" encrypted]")) {
            return;
        }

        String groupName = oldBook.getLore().get(0).substring(3).split(" encrypted")[0];

        //Permission check
        if (!player.hasPermission("bookswithoutborders.decrypt." + groupName) &&
                !(BooksWithoutBorders.adminDecrypt && player.hasPermission("bookswithoutborders.admin"))) {
            return;
        }

        String fileName = oldBook.getTitle() + BooksWithoutBorders.titleAuthorSeparator + oldBook.getAuthor();

        String encryptionFile = BooksWithoutBorders.bwb.cleanString(groupName) + slash + fileName + ".yml";

        File file = new File(bookFolder + "Encrypted" + slash + encryptionFile);
        if (!file.isFile()) {
            file = new File(bookFolder + fileName + ".txt");
            if (!file.isFile()) {
                return;
            }
        }
        newBook = BooksWithoutBorders.bwb.loadBook(player, fileName, "true", groupName, heldItem.getAmount());

        player.getInventory().setItem(hand, newBook);
        player.closeInventory();
        player.sendMessage(ChatColor.GREEN + "Book auto-decrypted!");
    }

    /**
     * Encrypts and replaces the player's used book
     *
     * @param sign         <p>The clicked sign</p>
     * @param heldItemType <p>The item type used to click the sign</p>
     * @param player       <p>The player which clicked the sign</p>
     * @param hand         <p>The EquipmentSlot of the used hand</p>
     */
    private void encryptHeldBookUsingSign(Sign sign, Material heldItemType, Player player, EquipmentSlot hand) {
        ItemStack eBook;
        if (heldItemType == Material.WRITTEN_BOOK) {
            player.closeInventory();
            eBook = BooksWithoutBorders.bwb.bookEncryption(player, sign.getLine(2).substring(2), sign.getLine(3).substring(2), "");
            if (eBook != null) {
                player.getInventory().setItem(hand, eBook);
            }
        }
    }

    /**
     * Gives a player the book specified on a sign
     *
     * @param sign   <p>The sign the user clicked</p>
     * @param player <p>The player which clicked the sign</p>
     */
    private void giveBook(Sign sign, Player player) {
        String fileName = sign.getLine(2).substring(2);
        boolean isLoadListNumber = false;

        for (int x = 0; x < fileName.length(); x++) {
            if (!Character.isDigit(fileName.charAt(x)))
                break;
            if (x == fileName.length() - 1) {
                BooksWithoutBorders.loadList.put(player.getName(), BooksWithoutBorders.bwb.listFiles(player, true, 0, true));
                isLoadListNumber = true;
            }
        }

        if (!isLoadListNumber && sign.getLine(3).length() >= 2)
            fileName = sign.getLine(2).substring(2) + sign.getLine(3).substring(2);

        ItemStack newBook = BooksWithoutBorders.bwb.loadBook(player, fileName, "true", "public");

        if (newBook != null) {
            player.getInventory().addItem(newBook);
            player.sendMessage(ChatColor.GREEN + "Received book!");
        } else {
            player.sendMessage(ChatColor.RED + "Book failed to load!");
        }
    }

    /**
     * Decrypts and replaces the player's used book
     *
     * @param sign         <p>The clicked sign</p>
     * @param heldItemType <p>The item type used to click the sign</p>
     * @param player       <p>The player which clicked the sign</p>
     * @param hand         <p>The EquipmentSlot of the used hand</p>
     */
    private void decryptHeldBookUsingSign(Sign sign, Material heldItemType, Player player, EquipmentSlot hand) {
        //Decrypt the held book and replace it
        if (heldItemType == Material.WRITTEN_BOOK) {
            player.closeInventory();

            //Converts user supplied key into integer form
            StringBuilder key = new StringBuilder();
            String lineText = ChatColor.stripColor(sign.getLine(2));
            for (int x = 0; x < lineText.length(); x++) {
                key.append(Character.getNumericValue(Character.codePointAt(lineText, x)));
            }

            ItemStack book = BooksWithoutBorders.bwb.eLoad(player, key.toString(), false);
            if (book != null) {
                player.getInventory().setItem(hand, book);
                player.sendMessage(ChatColor.GREEN + "Book decrypted!");
            }
        }
    }

    /**
     * Gets the color of a line on a sign
     *
     * @param sign       <p>The sign to check</p>
     * @param lineNumber <p>The line number to check</p>
     * @return <p>The color of the sign</p>
     */
    private ChatColor getSignLineColor(Sign sign, int lineNumber) {
        return ChatColor.getByChar(sign.getLine(lineNumber).substring(0, 2));
    }

    /**
     * Checks if a line on a sign equals some string, and that the sign is the correct color
     *
     * @param sign       <p>The sign to read</p>
     * @param lineNumber <p>The sign line to read</p>
     * @param compareTo  <p>The string to look for</p>
     * @param color      <p>The color to match</p>
     * @return <p>True if the given string is what's on the sign</p>
     */
    private boolean signLineEquals(Sign sign, int lineNumber, String compareTo, ChatColor color) {
        String line = sign.getLine(lineNumber);
        return line.equalsIgnoreCase(color + compareTo);
    }

}
