package net.knarcraft.bookswithoutborders;

import java.math.BigInteger;
import java.util.StringTokenizer;

public class SubstitutionCipher {

    public SubstitutionCipher() {

    }

    // encrypts a string using a substitution cipher.
    // the substitution is made harder to crack by
    // using a string for the key, it is converted
    // a series of offsets that each character in the
    // original message is offset by
    public String encrypt(String in, String key) {
        StringBuilder output = new StringBuilder();
        if (key != null && key.length() > 0) {
            StringTokenizer tokenizer = new StringTokenizer(key, ", "); // tokenizes the key
            // converts each number in the key to an integer and adds to an array
            int[] offsetArray = new int[tokenizer.countTokens()];
            for (int i = 0; i < offsetArray.length; i++) {
                String nt = tokenizer.nextToken();

                try {
                    offsetArray[i] = Integer.parseInt(nt);
                } catch (NumberFormatException e) {
                    BigInteger big = new BigInteger(nt);
                    offsetArray[i] = Math.abs(big.intValue());
                }
            }

            int offsetPosition = 0;
            for (int i = 0; i < in.length(); i++) {
                output.append((char) (in.charAt(i) + offsetArray[offsetPosition])); //encrypts the letter and adds to the output string
                // uses the next offset in the key, goes back to first offset if at end of list
                if (offsetPosition < offsetArray.length - 1) {
                    offsetPosition++;
                } else {
                    offsetPosition = 0;
                }
            }
        }
        return output.toString();
    }

    // decrypts a string using the same substitution method,
    // but in reverse. Could probably be combined into one
    // method with a flag for encryption / decryption, but
    // I'm lazy.
    public String decrypt(String in, String key) {
        StringBuilder output = new StringBuilder();
        if (key != null && key.length() > 0) {
            StringTokenizer tokenizer = new StringTokenizer(key, ", "); // tokenizes the key
            // converts each number in the key to an integer and adds to an array
            int[] offsetArray = new int[tokenizer.countTokens()];
            for (int i = 0; i < offsetArray.length; i++) {
                offsetArray[i] = Integer.parseInt(tokenizer.nextToken());
            }
            int offsetPosition = 0;
            for (int i = 0; i < in.length(); i++) {
                output.append((char) (in.charAt(i) - offsetArray[offsetPosition])); //encrypts the letter and adds to the output string
                // uses the next offset in the key, goes back to first offset if at end of list
                if (offsetPosition < offsetArray.length - 1) {
                    offsetPosition++;
                } else {
                    offsetPosition = 0;
                }
            }
        }
        return output.toString();
    }


    // the one time pad encryption is very secure, and
    // encryption works just like decryption, but is
    // vulnerable if the same key is used more than once.
    public String oneTimePad(String in, String key) {
        StringBuilder output = new StringBuilder();
        for (int i = 0; i < in.length(); i++) {
            output.append((char) (in.charAt(i) ^ key.charAt(i % key.length())));
        }
        return output.toString();
    }
}

